%% T-61.2010 Datasta tietoon / T#3 / 1
% Jukka Parviainen, 17.11.2006

clear;         % clear memory of variables
clc;           % clear the command window
close all;     % close all windows

%% k-nearest neighbour classifier
% Let's take a basic example from problem 3 of paper excercise
% round #3
%
% Suppose we have a set of data points (e.g. measurements).
% Suppose we know for each data point whether it belongs to
% class A or class B (a simple two-class scenario).
% Let the dimension of each data point be d (i.e., each data point
% consists of d measured values)
%
% Let's construct a classifier that let's us classify novel data
% points to belong to either class A or class B.
%
% Let our labeled data set consist of 5 samples of class A 
% (visualised as circles) and and 4 samples of class B (square).
% The data dimension is 2, thus the points can be drawn in
% xy-plane.

A = [ 0 1 3    5    8.5; ...
    0 3 1.8  -0.8 3.3]
  
B = [ 5.2  8   9   12; ...
    3.5  -3  0.5 2.5]

figure(11); clf;
pA = plot(A(1,:), A(2,:), 'o');
hold on;
pB = plot(B(1,:), B(2,:), 's');
axis equal
grid on;
axis([-2.2 14.5 -5.2 5.8]);
set(gca,'XTick',[-10:1:30],'YTick',[-10:1:20]);
legend({'class A','class B'});
title('Training data for classes A and B');

%%  Training and use of the classifier
% These labeled points form the "training set".
% A novel data point is now classified to belong to the class
% to which majority of its k neighbours belong.
%
% Let's generate N new points and see to which of the classes they belong.

N = 5;
C = [14*rand(1,N)-2; 8*rand(1,N)-4];

disp('To which class (A or B) the following data points belong?');
disp(C);

for n = [1 : N]
    % n:th value from C
    plo1(n) = plot(C(1,n), C(2,n), 'r*');
    tex1(n) = text(C(1,n)+0.2, C(2,n), [num2str(n)]);
    tex2(n) = text(C(1,n), C(2,n)+0.7, '?');
    set([tex1 tex2],'Color','r','FontSize',16);
end;
title('Training set and the points to be classified');
shg;

%% Example: the distances between the point C(1) and all the
% training points
% The distance is calculated as Euclidean distance
%
%$$ dist = \sqrt{ (c_1 - x_1)^2 + (c_2 - x_2)^2 } $$
%

distC1 = zeros(1,size(A,2)+size(B,2));
for k = [1 : size(A,2)]
    distC1(k) = sqrt( (C(1,1)-A(1,k))^2 + (C(2,1)-A(2,k))^2 );
end;
for k = [1 : size(B,2)]
    distC1(k+size(A,2)) = sqrt( (C(1,1)-B(1,k))^2 + (C(2,1)-B(2,k))^2 );
end;

disp('Distances from the point C(1)');
disp(C(:,1));
disp('to the points of classes A and B:');
disp(distC1);

%% Plot to a diagram
% (here merely plotting commands)

lin6 = zeros(size(A,2)+size(B,2),1);
for k = [1 : size(A,2)]
    lin6(k) = line([C(1,1) A(1,k)],[C(2,1) A(2,k)],'Color',[0 0 0]);
end;
for k = [1 : size(B,2)]
    lin6(k+size(A,2)) = line([C(1,1) B(1,k)],[C(2,1) B(2,k)],'Color',[0 0 0]);
end;
[distx, indx] = max(distC1);
[distn, indn] = min(distC1);
set(lin6(indn),'LineWidth',3);
tex6(1) = text(C(1,1)-1,C(2,1)+0.7,['min dist = ' num2str(distn)]);
tex6(2) = text(C(1,1)-1,C(2,1)+1.4,['max dist = ' num2str(distx)]);
title('Training set, the points to be classified and the distances from the point C(1)');
shg;

%% I.e. classifying the point according to its nearest neighbour!
% (once again just plotting commands)

% First remove some things plotted in red
delete(lin6);
delete(tex6);
delete(tex1(1));
delete(tex2(1));
delete(plo1(1));

if indn <= size(A,2)
    plot(C(1,1), C(2,1), 'ro');
else
    plot(C(1,1), C(2,1), 'rs');
end;
tex7 = text(C(1,1), C(2,1)+0.3, num2str(1), 'Color',[1 0 0]);
shg;

%% Calculate the distance from all the novel points to the trainig
% points using the Euclidean distance
% 

dist = pdist([A B C].');
% lower triangle in the representation squareform(dist)

sqDist = squareform(dist);
% symmetric distance matrix

ABDist = sqDist(10:end, 1:(size(A,2)+size(B,2)));
% take just the part of the distance matrix that contains 
% the distances to all the points in A and all the points in B


disp('On each row the distance of the points in C from');
disp('all the traing samples in A and B');
disp(['A: columns 1..' num2str(size(A,2)) ', B: ' num2str(size(A,2)+1) '..' num2str(size(A,2)+size(B,2))]);
disp('-----------------------------------------------------');
disp(ABDist)
disp('-----------------------------------------------------');

%% Plot the novel points with their class labels into a graph
% Simultaneously find the nearest neighbour

% Erase things plotted in red
delete(tex1(2:end));
delete(tex2(2:end));
delete(plo1(2:end));
delete(tex7);

for n = [1 : N]
    % n:th value from C
    [minv, minind] = min(ABDist(n, :));
    if (minind <= size(A,2))
        % closest neighbour is from class A
        plot(C(1,n), C(2,n), 'ro');
        t1 = text(C(1,n)+0.2, C(2,n), [num2str(n)]);
        set(t1, 'Color', 'r');
        Cclass(n) = 'A';
    else
        plot(C(1,n), C(2,n), 'rs');
        t1 = text(C(1,n)+0.2, C(2,n), [num2str(n)]);
        set(t1, 'Color', 'r');
        Cclass(n) = 'B';
    end;
end;
title('Training set and the classified points');

disp('Class labels of the novel points C according to a 1-NN classifier:');
disp(Cclass);
shg;

%% Voronoi regions
% kNN-classifier is directly based on the training set
% a test point is assigned to the class that most of its 
% k neihbours belong to
%
% in the 1-NN case the decision surface (boundary) can be drawn:

[vorx, vory] = voronoi([A(1,:) B(1,:)],[A(2,:) B(2,:)]);
hold on;
pm = plot(vorx, vory, 'b:');
% The following line works only in newest versions of Matlab:
%pn = plot(vorx(:,[16 5 6 7 8 15 9 18]), vory(:,[16 5 6 7 8 15 9 18]));
%set(pn, 'LineWidth', 3,'Color', [0 0 1]);
title('The training set, classified points and Voronoi regions');
shg;

%% Demonstrations of kNN in the net
% For example:
% |http://www.cs.cmu.edu/~zhuxj/courseproject/knndemo/KNN.html|
% that demonstrates what can happen if the training set is chosen poorly:
%
% * "create truth": the true class boundaries appear on green-white in the 
%   window on the left
% * "generate samples": plot random points of the training set 
%   to the leftmost graph
% * "classify": create decision boundaries on the basis of 
%   the training set to the graph on the right. Determine the 
%   classification error, i.e. measure how close the green-white regions
%   are to each other
%
