%% T-61.2010 Datasta tietoon / T#3 / 2
% Jukka Parviainen, 21.11.2006

%clear;         % clear all the variables from memory
%clc;           % clear the command window
%close all;     % close all graph windows

%clear ind_data;  % if you want to change random sampling

%% kNN classifier
kNN = 1;  % choose either 1, 3 or 5. We begin with 1.
disp('********************************************************');
disp(['kNN-classifier: choose ' num2str(kNN) ' nearest neighbours']);

%% Training and test set
% Usually we have labeled data set in our disposal
                              
% Partition it into two: trainig and test sets
%
% Usually the classifier is trained with training data,
% and the test data is used to evaluate the quality of the produced
% classifier by calculating classification error.
% After this the classifier is applied to completely novel
% data whose class labels are not known but determined by the classifier

%
% Load the data set and class labels

load dt_t3_m2.mat
disp('Read variables "data" ja "dclass"');
disp('Each 2D point is associated with a class label +1/-1, for example');
disp(data(:, [1 70 130]));
disp(dclass([1 70 130]));

%% Partition the data into training and test sets
% Randomly choose half of the data for training.
% 
% Class label either -1 or +1.

N = size(data, 2);
if (exist('ind_data') ~= 1)
    %created only once! 
    % If you want to randomly select new set of points, execute command "clear ind_data"
    ind_data = randperm(N);
end;

train = data(:, ind_data(1 : N/2));
test  = data(:, ind_data(N/2 + 1 : end));
classtrain = dclass(ind_data(1 : N/2));
classtest  = dclass(ind_data(N/2 + 1 :end));
N_train_A = sum(classtrain == 1);
N_test_A = sum(classtest == 1);
disp(['Now the training set contains ' num2str(N_train_A) ' examples of "+1" and ' num2str(N/2-N_train_A) ' examples of "-1"']);
disp(['and the test set of ' num2str(N_test_A) ' "+1" and ' num2str(N/2-N_test_A) ' "-1"']);

%% Plot the data
% Blue crosses denote training samples with label +1,
  % blue circles training samples with class -1,
% and black diamonds the test samples
% whose class labels are known but utilised 
% only when calculating the error produced by the classifier.

ind_A = find(classtrain > 0);
ind_B = find(classtrain < 0);

figure(20+kNN); clf;
plot(train(1,ind_A), train(2, ind_A), 'bx');
hold on;
plot(train(1,ind_B), train(2, ind_B), 'bo');
pl1 = plot(test(1,:), test(2,:), 'kd');
title('Training data in two classes and the test data');
legend({'Class +1 (train)','Class -1 (train)','The points to be classified (test)'});
axis equal;
grid on;

%% Calculate the matrix of distances from the training samples
DistVect   = pdist([train test]');
DistMatrix = squareform(DistVect);
Dist       = DistMatrix(N/2 + 1: end, 1:N/2);
% Here first row contains distances from the first test point to all the 
% training points

%% Form the k-NN classifier
class_correct   = 0;
class_incorrect = 0;
class_A = [];
class_B = [];
ind_incorrect = [];
class_A_was_B = [];
class_B_was_A = [];
 
for m = 1 : N/2
    [mins, ind_mins] = sort(Dist(m, :));
    classified = sum(dclass(ind_data(ind_mins(1:kNN))));
    if (sign(dclass(ind_data(m + N/2))) == sign(classified))
        class_correct = class_correct + 1;
        if (classified > 0)
            class_A = [class_A m];
        else
            class_B = [class_B m];
        end         
    else
        class_incorrect = class_incorrect + 1;
        ind_incorrect = [ind_incorrect m];
        if (classified < 0)
            class_A_was_B = [class_A_was_B m];
        else
            class_B_was_A = [class_B_was_A m];
        end         
    end;
end;
error = 100 * class_incorrect / (N/2);
disp(['Outcome of the classification ' num2str(kNN) '-NN:']);
disp(['  correct:  ' num2str(class_correct)]);
disp(['  incorrect:  ' num2str(class_incorrect)]);
disp(['Classification error: ' num2str(error) '%']);        

if (exist('pl1') == 1), delete(pl1); end;
plot(test(1,class_A), test(2, class_A), 'gx');
plot(test(1,class_B), test(2, class_B), 'go');
plot(test(1,class_A_was_B), test(2, class_A_was_B), 'rx');
plot(test(1,class_B_was_A), test(2, class_B_was_A), 'ro');

title([num2str(kNN) '-NN: training data in twoclasses, test data by correctness of classification']);
legend({'Class +1 (train)','Class -1 (train)', ...
    ['Correct +1 [' num2str(length(class_A)) ']'], ...
    ['Correct -1 [' num2str(length(class_B)) ']'], ...
    ['Incorrect +1 [' num2str(length(class_A_was_B)) ']'], ...
    ['Incorrect -1 [' num2str(length(class_B_was_A)) ']']});
shg;

%% Comparison, comments
% Classification errors by 1-NN, 3-NN, 5-NN?
%
% Advantages / disadvantages? Differences between 1, 3, 5?

